#import os
# from foo import junk
#a = 3
a = 4
#foo(1, 2, 3)

def foo(x, y, z):
    content = 1 # print('hello')
    print(x, y, z)

    # This is a real comment.
    #return True
    return False

#import os  # noqa: ERA001



# module members cannot be imported with that syntax
import typing.TypedDict

# we don't track reassignments
import typing, other
typing = other
typing.TypedDict()

# yet another false positive
def foo(typing):
    typing.TypedDict()



# These remain unchanged
str(1)
str(*a)
str("foo", *a)
str(**k)
str("foo", **k)
str("foo", encoding="UTF-8")
str("foo"
    "bar")
bytes("foo", encoding="UTF-8")
bytes(*a)
bytes("foo", *a)
bytes("foo", **a)
bytes(b"foo"
      b"bar")

# These become string or byte literals
str()
str("foo")
str("""
foo""")
bytes()
bytes(b"foo")
bytes(b"""
foo""")



# These should not change
raise ValueError
raise ValueError(1)

from .mmap import error
raise error

# Testing the modules
import socket, mmap, select
raise socket.error
raise mmap.error
raise select.error

raise socket.error()
raise mmap.error(1)
raise select.error(1, 2)

raise socket.error(
    1,
    2,
    3,
)

from mmap import error
raise error

from socket import error
raise error(1)

from select import error
raise error(1, 2)

# Testing the names
raise EnvironmentError
raise IOError
raise WindowsError

raise EnvironmentError()
raise IOError(1)
raise WindowsError(1, 2)

raise EnvironmentError(
    1,
    2,
    3,
)

raise WindowsError
raise EnvironmentError(1)
raise IOError(1, 2)



import sys
import tempfile

print("Hello, world!")  # T201
print("Hello, world!", file=None)  # T201
print("Hello, world!", file=sys.stdout)  # T201
print("Hello, world!", file=sys.stderr)  # T201

with tempfile.NamedTemporaryFile() as fp:
    print("Hello, world!", file=fp)  # OK



import mmap, select, socket
from mmap import error
# These should be fixed
try:
    pass
except EnvironmentError:
    pass

try:
    pass
except IOError:
    pass

try:
    pass
except WindowsError:
    pass

try:
    pass
except mmap.error:
    pass

try:
    pass
except select.error:
    pass

try:
    pass
except socket.error:
    pass

try:
    pass
except error:
    pass

# Should NOT be in parentheses when replaced

try:
    pass
except (IOError,):
    pass
try:
    pass
except (mmap.error,):
    pass
try:
    pass
except (EnvironmentError, IOError, OSError, select.error):
    pass

# Should be kept in parentheses (because multiple)

try:
    pass
except (IOError, KeyError, OSError):
    pass

# First should change, second should not
from .mmap import error
try:
    pass
except (IOError, error):
    pass
# These should not change

from foo import error

try:
    pass
except (OSError, error):
    pass

try:
    pass
except:
    pass

try:
    pass
except AssertionError:
    pass
try:
    pass
except (mmap).error:
    pass

try:
    pass
except OSError:
    pass

try:
    pass
except (OSError, KeyError):
    pass



"""Test: Mypy extensions."""

from mypy_extensions import DefaultNamedArg

# OK
_ = DefaultNamedArg(bool | None, name="some_prop_name")
_ = DefaultNamedArg(type=bool | None, name="some_prop_name")
_ = DefaultNamedArg(bool | None, "some_prop_name")

# Not OK
_ = DefaultNamedArg("Undefined", name="some_prop_name")
_ = DefaultNamedArg(type="Undefined", name="some_prop_name")
_ = DefaultNamedArg("Undefined", "some_prop_name")



# This should ignore both errors.
from typing import (  # noqa: F401
    List,
    Sequence,
)

# This should ignore both errors.
from typing import (  # noqa
    List,
    Sequence,
)

# This should ignore both errors.
from typing import (
    List,  # noqa: F401
    Sequence,  # noqa: F401
)

# This should ignore both errors.
from typing import (
    List,  # noqa
    Sequence,  # noqa
)

# This should ignore the first error.
from typing import (
    Mapping,  # noqa: F401
    Union,
)

# This should ignore both errors.
from typing import (  # noqa
    List,
    Sequence,
)

# This should ignore the error, but the inner noqa should be marked as unused.
from typing import (  # noqa: F401
    Optional,  # noqa: F401
)

# This should ignore the error, but the inner noqa should be marked as unused.
from typing import (  # noqa
    Optional,  # noqa: F401
)

# This should ignore the error, but mark F501 as unused.
from typing import (  # noqa: F401
    Dict,  # noqa: F501
)

# This should ignore the error, but mark F501 as unused.
from typing import (  # noqa: F501
    Tuple,  # noqa: F401
)

# This should ignore both errors.
from typing import Any, AnyStr  # noqa: F401

# This should ignore both errors.
from typing import AsyncIterable, AsyncGenerator  # noqa

# This should mark F501 as unused.
from typing import Awaitable, AwaitableGenerator  # noqa: F501



import sys, math
from os import path, uname
from logging.handlers import StreamHandler, FileHandler

# comment 1
from third_party import lib1, lib2, \
     lib3, lib7, lib5, lib6
# comment 2
from third_party import lib4

from foo import bar  # comment 3
from foo2 import bar2  # comment 4

# comment 5
from bar import (
     a, # comment 6
     b, # comment 7
)


from __future__ import annotations


def f_a():
    raise RuntimeError("This is an example exception")


def f_a_short():
    raise RuntimeError("Error")


def f_b():
    example = "example"
    raise RuntimeError(f"This is an {example} exception")


def f_c():
    raise RuntimeError("This is an {example} exception".format(example="example"))


def f_ok():
    msg = "hello"
    raise RuntimeError(msg)



import datetime
import datetime as dt
from datetime import timezone
from datetime import timezone as tz

print(datetime.timezone(-1))
print(timezone.utc)
print(tz.utc)

print(datetime.timezone.utc)
print(dt.timezone.utc)



import datetime

# no args
datetime.datetime.fromtimestamp(1234)

# wrong keywords
datetime.datetime.fromtimestamp(1234, bad=datetime.timezone.utc)

# none args
datetime.datetime.fromtimestamp(1234, None)

# none keywords
datetime.datetime.fromtimestamp(1234, tz=None)

from datetime import datetime

# no args unqualified
datetime.fromtimestamp(1234)



###
# Errors.
###
def f():
    print(x)

    global x

    print(x)


def f():
    global x

    print(x)

    global x

    print(x)


def f():
    print(x)

    global x, y

    print(x)


def f():
    global x, y

    print(x)

    global x, y

    print(x)


def f():
    x = 1

    global x

    x = 1


def f():
    global x

    x = 1

    global x

    x = 1


def f():
    del x

    global x, y

    del x


def f():
    global x, y

    del x

    global x, y

    del x


def f():
    del x

    global x

    del x


def f():
    global x

    del x

    global x

    del x


def f():
    del x

    global x, y

    del x


def f():
    global x, y

    del x

    global x, y

    del x


def f():
    print(f"{x=}")
    global x


###
# Non-errors.
###
def f():
    global x

    print(x)


def f():
    global x, y

    print(x)


def f():
    global x

    x = 1


def f():
    global x, y

    x = 1


def f():
    global x

    del x


def f():
    global x, y

    del x


def f():
    global x
    print(f"{x=}")



import datetime

# no args
datetime.datetime.now()

# wrong keywords
datetime.datetime.now(bad=datetime.timezone.utc)

# none args
datetime.datetime.now(None)

# none keywords
datetime.datetime.now(tz=None)

from datetime import datetime

# no args unqualified
datetime.now()



import subprocess
import subprocess as somename
from subprocess import run
from subprocess import run as anothername

subprocess.run(["foo"], universal_newlines=True, check=True)
somename.run(["foo"], universal_newlines=True)

run(["foo"], universal_newlines=True, check=False)
anothername(["foo"], universal_newlines=True)

subprocess.run(["foo"], check=True)



import mmap, socket, select

try:
    pass
except (OSError, mmap.error, IOError):
    pass
except (OSError, socket.error, KeyError):
    pass

try:
    pass
except (
    OSError,
    select.error,
    IOError,
):
    pass



import datetime

# no args
datetime.datetime(2000, 1, 1, 0, 0, 0)

# none args
datetime.datetime(2000, 1, 1, 0, 0, 0, 0, None)

# not none arg
datetime.datetime(2000, 1, 1, 0, 0, 0, 0, datetime.timezone.utc)

# no kwargs
datetime.datetime(2000, 1, 1, fold=1)

# none kwargs
datetime.datetime(2000, 1, 1, tzinfo=None)

from datetime import datetime

# no args unqualified
datetime(2000, 1, 1, 0, 0, 0)



## Banned modules ##
import cgi

from cgi import *

from cgi import a, b, c

# banning a module also bans any submodules
import cgi.foo.bar

from cgi.foo import bar

from cgi.foo.bar import *

## Banned module members ##

from typing import TypedDict

import typing

# attribute access is checked
typing.TypedDict

typing.TypedDict.anything

# function calls are checked
typing.TypedDict()

typing.TypedDict.anything()

# import aliases are resolved
import typing as totally_not_typing
totally_not_typing.TypedDict



# These should change
x = u"Hello"

u'world'

print(u"Hello")

print(u'world')

import foo

foo(u"Hello", U"world", a=u"Hello", b=u"world")

# These should stay quoted they way they are

x = u'hello'
x = u"""hello"""
x = u'''hello'''
x = u'Hello "World"'

# These should not change
u = "Hello"

u = u

def hello():
    return"Hello"



import logging
from logging import warn

logging.warn("this is not ok")
log.warn("this is also not ok")
warn("not ok")


def foo():
    from logging import warn

    def warn():
        pass

    warn("has been redefined, but we will still report it")



def f():
    "Here's a line without a period"
    ...


def f():
    """Here's a line without a period"""
    ...


def f():
    """
    Here's a line without a period,
    but here's the next line
    """
    ...


def f():
    """Here's a line without a period"""
    ...


def f():
    """
    Here's a line without a period,
    but here's the next line"""
    ...


def f():
    """
    Here's a line without a period,
    but here's the next line with trailing space """
    ...


def f():
    r"Here's a line without a period"
    ...


def f():
    r"""Here's a line without a period"""
    ...


def f():
    r"""
    Here's a line without a period,
    but here's the next line
    """
    ...


def f():
    r"""Here's a line without a period"""
    ...


def f():
    r"""
    Here's a line without a period,
    but here's the next line"""
    ...


def f():
    r"""
    Here's a line without a period,
    but here's the next line with trailing space """
    ...


def f(rounds: list[int], number: int) -> bool:
    """
    :param rounds: list - rounds played.
    :param number: int - round number.
    :return:  bool - was the round played?
    """
    return number in rounds


def f(rounds: list[int], number: int) -> bool:
    """
    Args:
        rounds (list): rounds played.
        number (int): round number.

    Returns:
        bool: was the round played?
    """
    return number in rounds



# These should be changed
if True:
    import mock

if True:
    import mock, sys

# This goes to from unitest import mock
import mock.mock

# Mock should go on a new line as `from unittest import mock`
import contextlib, mock, sys

# Mock should go on a new line as `from unittest import mock`
import mock, sys
x = "This code should be preserved one line below the mock"

# Mock should go on a new line as `from unittest import mock`
from mock import mock

# Should keep trailing comma
from mock import (
    mock,
    a,
    b,
    c,
)

# Should not get a trailing comma
from mock import (
    mock,
    a,
    b,
    c
)

if True:
    if False:
        from mock import (
            mock,
            a,
            b,
            c
        )

# These should not change:
import os, io

# Mock should go on a new line as `from unittest import mock`
import mock, mock

# Mock should go on a new line as `from unittest import mock as foo`
import mock as foo

# Mock should go on a new line as `from unittest import mock as foo`
from mock import mock as foo

if True:
    # This should yield multiple, aliased imports.
    import mock as foo, mock as bar, mock

    # This should yield multiple, aliased imports, and preserve `os`.
    import mock as foo, mock as bar, mock, os

if True:
    # This should yield multiple, aliased imports.
    from mock import mock as foo, mock as bar, mock


# This should be unchanged.
x = mock.Mock()

# This should change to `mock.Mock`().
x = mock.mock.Mock()



# This has a magic trailing comma, will be sorted, but not rolled into one line
from sys import (
    stderr,
    argv,
    stdout,
    exit,
)

# No magic comma, this will be rolled into one line.
from os import (
    path,
    environ,
    execl,
    execv
)

from glob import (
    glob,
    iglob,
    escape,  # Ends with a comment, should still treat as magic trailing comma.
)

# These will be combined, but without a trailing comma.
from foo import bar
from foo import baz

# These will be combined, _with_ a trailing comma.
from module1 import member1
from module1 import (
    member2,
    member3,
)

# These will be combined, _with_ a trailing comma.
from module2 import member1, member2
from module2 import (
    member3,
)



###
# Errors
###
if "abc" is "def":  # F632 (fix)
    pass
if "abc" is None:  # F632 (fix, but leaves behind unfixable E711)
    pass
if None is "abc":  # F632 (fix, but leaves behind unfixable E711)
    pass
if "abc" is False:  # F632 (fix, but leaves behind unfixable E712)
    pass
if False is "abc":  # F632 (fix, but leaves behind unfixable E712)
    pass
if False == None:  # E711, E712 (fix)
    pass
if None == False:  # E711, E712 (fix)
    pass

###
# Unfixable errors
###
if "abc" == None:  # E711
    pass
if None == "abc":  # E711
    pass
if "abc" == False:  # E712
    pass
if False == "abc":  # E712
    pass

###
# Non-errors
###
if "def" == "abc":
    pass
if False is None:
    pass
if None is False:
    pass



###
# Errors.
###
def f():
    global x


def f():
    global x

    print(x)


###
# Non-errors.
###
def f():
    global x

    x = 1


def f():
    global x

    (x, y) = (1, 2)


def f():
    global x

    del x



def f(x, y, z):
    """Do f.

    Args:
        x: the value
            with a hanging indent

    Returns:
        the value
    """
    return x


def f(x, y, z):
    """Do f.

    Args:
        x:
            The whole thing has a hanging indent.

    Returns:
        the value
    """
    return x


def f(x, y, z):
    """Do f.

    Args:
        x:
            The whole thing has a hanging indent.

    Returns: the value
    """
    return x


def f(x, y, z):
    """Do f.

    Args:
        x: the value def
            ghi

    Returns:
        the value
    """
    return x


def f(x, y, z):
    """Do f.

    Args:
        x: the value
        z: A final argument

    Returns:
        the value
    """
    return x


def f(x, y, z):
    """Do g.

    Args:
        x: the value
        z: A final argument

    Returns: the value
    """
    return x


def f(x, y, z):
    """Do h.

    Args:
        x: the value
        z: A final argument
    """
    return x



from subprocess import run
import subprocess

output = run(["foo"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

output = subprocess.run(["foo"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

output = subprocess.run(stdout=subprocess.PIPE, args=["foo"],  stderr=subprocess.PIPE)

output = subprocess.run(
    ["foo"], stdout=subprocess.PIPE, check=True, stderr=subprocess.PIPE
)

output = subprocess.run(
    ["foo"], stderr=subprocess.PIPE, check=True, stdout=subprocess.PIPE
)

output = subprocess.run(
    ["foo"],
    stdout=subprocess.PIPE,
    check=True,
    stderr=subprocess.PIPE,
    text=True,
    encoding="utf-8",
    close_fds=True,
)

if output:
    output = subprocess.run(
        ["foo"],
        stdout=subprocess.PIPE,
        check=True,
        stderr=subprocess.PIPE,
        text=True,
        encoding="utf-8",
    )


# Examples that should NOT trigger the rule
from foo import PIPE
subprocess.run(["foo"], stdout=PIPE, stderr=PIPE)
run(["foo"], stdout=None, stderr=PIPE)



# These two imports have something after cElementTree, so they should be fixed.
from xml.etree.cElementTree import XML, Element, SubElement
import xml.etree.cElementTree as ET

# Weird spacing should not cause issues.
from   xml.etree.cElementTree    import  XML
import    xml.etree.cElementTree       as      ET

# Multi line imports should also work fine.
from xml.etree.cElementTree import (
    XML,
    Element,
    SubElement,
)
if True:
    import xml.etree.cElementTree as ET
    from xml.etree import cElementTree as CET

from xml.etree import cElementTree as ET

import contextlib, xml.etree.cElementTree as ET

# This should fix the second, but not the first invocation.
import xml.etree.cElementTree, xml.etree.cElementTree as ET

# The below items should NOT be changed.
import xml.etree.cElementTree

from .xml.etree.cElementTree import XML

from xml.etree import cElementTree



import datetime

# bad format
datetime.datetime.strptime("something", "%H:%M:%S%Z")

# no replace or astimezone
datetime.datetime.strptime("something", "something")

# wrong replace
datetime.datetime.strptime("something", "something").replace(hour=1)

# none replace
datetime.datetime.strptime("something", "something").replace(tzinfo=None)

# OK
datetime.datetime.strptime("something", "something").replace(
    tzinfo=datetime.timezone.utc
)

# OK
datetime.datetime.strptime("something", "something").astimezone()

# OK
datetime.datetime.strptime("something", "%H:%M:%S%z")

# OK
datetime.datetime.strptime("something", something).astimezone()

# OK
datetime.datetime.strptime("something", something).replace(tzinfo=datetime.timezone.utc)

from datetime import datetime

# no replace orastimezone unqualified
datetime.strptime("something", "something")



import typing
import typing as Hello
from typing import Text
from typing import Text as Goodbye


def print_word(word: Text) -> None:
    print(word)


def print_second_word(word: typing.Text) -> None:
    print(word)


def print_third_word(word: Hello.Text) -> None:
    print(word)


def print_fourth_word(word: Goodbye) -> None:
    print(word)
